#include "anircd.h"

int main(int ac, char **av)
{
	int		ls, maxfd, optval = 1, n;
	struct		sockaddr_in sin;
	time_t		mclr;

	/* Ignore SIGPIPE */
	signal(SIGPIPE, SIG_IGN);
	/* For rehashing */
	signal(SIGUSR1, rehash_motd);

	if (!log_init())
	{
		fprintf(stderr, "Failed to open log file; aborting.\n");
		return(-1);
	}

	if (!config_load(CONFIG_FILE))
	{
		fprintf(stderr, "Failed to open config file '%s'; aborting.\n", CONFIG_FILE);
		return(-1);
	}

	daemon(1, 0);

	ls = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (ls < 0)
	{
		perror("socket");
		return(-1);
	}

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = PF_INET;
	sin.sin_port = htons(c_port);
	sin.sin_addr.s_addr = INADDR_ANY;
#ifdef __FreeBSD__
	sin.sin_len = sizeof(sin);
#endif

	if (setsockopt(ls, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval)) < 0)
	{
		perror("setsockopt");
		exit(-1);
	}

	if (bind(ls, (struct sockaddr *) &sin, sizeof(sin)) < 0)
	{
		perror("bind");
		return(-1);
	}

	if (listen(ls, 5) < 0)
	{
		perror("listen");
		return(-1);
	}

	sock_nonblock(ls);

#ifdef SSL
	if (!init_SSL(c_sslkey))
		return(-1);
#endif

	load_motd();
	mclr = time(NULL);

	for (;;)
	{
		User	*u, *un;
		struct	timeval tv;
		fd_set	rfds, wfds;

		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_SET(ls, &rfds);

		memset(&tv, 0, sizeof(tv));
		tv.tv_sec = 0;
		tv.tv_usec = 50000;

		for (maxfd = ls, u = ulist; u; u = u->next)
		{
			maxfd = (u->s > maxfd) ? u->s : maxfd;
			FD_SET(u->s, &rfds);
/*			FD_SET(u->s, &wfds);*/
		}
		if (select(maxfd + 1, &rfds, &wfds, NULL, &tv) < 0)
		{
			if (errno == EINTR)
				continue;
			else
			{
				log_error("select() returned an error: %s, exiting", strerror(errno));
				break;
			}
		}

		if (FD_ISSET(ls, &rfds))
		{
			int	cs;
			struct	sockaddr_in	sin;
			socklen_t	addrlen = sizeof(sin);

			if ((cs = accept(ls, (struct sockaddr *) &sin, &addrlen)) < 0)
				log_error("accept() failed for client: %s", strerror(errno));
			else if (!sock_nonblock(cs))
				close(cs);
			else if (tor_detect(sin.sin_addr))
			{
				char	errbuf[256];

				log_error("Tor detected on IP %s", inet_ntoa(sin.sin_addr));
				snprintf(errbuf, sizeof(errbuf) - 1, ":%s NOTICE AUTH :Your client is running Tor. Tor is prohibited in order to limit ban evasion et al.\r\n", c_sname);
				write(cs, errbuf, strlen(errbuf));
				close(cs);
			}
			else if (!(u = user_new(cs)))
				close(cs);
#ifdef SSL
			else
			{
				int	optval = 1, r;

				BIO_socket_ioctl(cs, FIONBIO, &optval);
				u->bio = BIO_new_socket(cs, BIO_NOCLOSE);

				u->ssl = SSL_new(g_SSLctx);
				SSL_set_bio(u->ssl, u->bio, u->bio);
			}
#endif
		}

		for (u = ulist; u; u = un)
		{
			/* in case struct gets deleted! */
			un = u->next;
			n = 0;
			if (FD_ISSET(u->s, &rfds))
				while ((n = process_read(u)))
				{
					if (n == 1)
					{
						if (process_inbuf(u) < 0)
						{
							n = -1;
							break;
						}
					}
					else
						break;
				}

/*			if (FD_ISSET(u->s, &wfds)) */
			if (n >= 0)
				process_write(u);
		}

		if (difftime(time(NULL), mclr) >= 15)
		{
			User *u;
			for (u = ulist; u; u = u->next)
				u->nmcnt = 0;
			mclr = time(NULL);
		}
	}

	free_motd();

	return(0);
}
