#include "anircd.h"

#ifdef SSL
static int process_ssl_accept(User *u)
{
	int	r, n;

	if ((r = SSL_accept(u->ssl)) <= 0)
	{
		if ((n = SSL_get_error(u->ssl, r)) != SSL_ERROR_WANT_READ)
		{
			log_error("SSL negotiation failed: %d,%d", n, errno);
			user_kill2(u, "SSL negotiation failed");
			return(-1);
		}
	}
	else
	{
		u->flags |= UFL_SSLDONE;
		send_cmd(u, NULL, "PING :BROKEN-mIRC-SSL");
	}

	return(0);
}
#endif

int process_read(User *u)
{
	int	slen, n = 0;
	unsigned char	*in;
	int	err;

#ifdef SSL
	if (!(u->flags & UFL_SSLDONE))
		return(process_ssl_accept(u));
#endif

	if (!u->inbuf)
	{
		u->inlen = 1024;
		if (!(u->inbuf = (unsigned char *) malloc(u->inlen)))
		{
			log_error("Failed to allocate memory for u->inbuf in %s!", __func__);
			return(0);
		}
		*u->inbuf = '\0';
	}

#ifdef SSL
	for (in = u->inbuf + strlen(u->inbuf); (n = SSL_read(u->ssl, in, 1)) > 0; )
#else
	for (in = u->inbuf + strlen(u->inbuf); (n = read(u->s, in, 1)) > 0; )
#endif
	{
		if (in >= (u->inbuf + u->inlen - 2) || *in == '\n')
			break;
		if (*in != '\r') ++in;
	}

	*++in = '\0';

#ifdef SSL
	if (((err = SSL_get_error(u->ssl, n)) && err != SSL_ERROR_WANT_READ) || !n)
#else
	if ((n < 0 && errno != EAGAIN && errno != EINTR) || !n)
#endif
	{
#ifdef SSL
		if (ERR_get_error() != 33619571)
		{
			user_kill2(u, "Socket error while reading socket");
			log_error("Error while reading from socket %d: %d", u->s, err);
#else
			user_kill2(u, strerror(errno));
			log_error("Error while reading from socket %d: %d[%s]", u->s, errno, strerror(errno));
			errno = 0;
#endif
/*			user_kill2(u, "Socket error while reading socket"); */
			return(-1);
#ifdef SSL
		}
#endif
	}

	return(in[-1] == '\n' || in >= (u->inbuf + u->inlen - 2));
}

int process_write(User *u)
{
	int	n = 0, r;

	if (!u->outbuf)
		return(1);

#ifdef SSL
	if ((n = SSL_write(u->ssl, u->outbuf, strlen(u->outbuf))) < 0 && (r = SSL_get_error(u->ssl, n)) != SSL_ERROR_WANT_READ && r != SSL_ERROR_WANT_WRITE)
#else
	if ((n = write(u->s, u->outbuf, strlen(u->outbuf))) < 0 && errno != EAGAIN && errno != EINTR)
#endif
	{
#ifdef SSL
		log_error("Error while writing to socket %d: %s, SSL:%d", u->s, strerror(errno), r);
#else
		log_error("Error while writing to socket %d: %d[%s]", u->s, errno, strerror(errno));
#endif
		return(0);
	}

	if (n > 0)
	{
		if (n < strlen(u->outbuf))
		{
			memmove(u->outbuf, u->outbuf + n, strlen(u->outbuf) - n + 1);
			if (!(u->outbuf = (unsigned char *) realloc(u->outbuf, strlen(u->outbuf) + 1)))
			{
				log_error("Failed to realloc() for u->outbuf in %s", __func__);
				return;
			}
		}
		else
		{
			free(u->outbuf);
			u->outbuf = NULL;
		}
	}

#ifdef SSL
	BIO_flush(u->bio);
#endif
	return(1);
}

int process_inbuf(User *u)
{
	int	ac, rc = 0;
	char	**av, *source, *cmd, *s, *t;
	char	*buf;

	strip(u->inbuf);
	buf = strdup(u->inbuf);

	for (ac = 0, av = NULL, s = ((*buf == ':') ? buf + 1 : buf); ; s = t)
	{
		av = (char **) realloc(av, sizeof(char *) * ++ac);
		av[ac - 1] = (*s != ':') ? s : s + 1;
		if ((t = strchr(s, ' ')) && *s != ':')
			*t = (char) 0;
		else
			break;
		while (isspace(*++t));
	}
	source = (*buf == ':') ? av[0] : NULL;
	cmd = source ? av[1] : av[0];
	ac -= source ? 2 : 1;
	av += source ? 2 : 1;

	strtoupper(cmd);

	if (source)
	{
		s = strchr(source, '!');
		if (s)
			*s = '\0';
	}

	if (u->nmcnt > MAX_MSG)
	{
		user_kill(u, "Excess flood");
		rc = -1;
		goto out;
	}

	/* 'source' is not used as the user context is already known */
	if (!(u->flags & UFL_REG))
	{
		if (!strcasecmp(cmd, "USER"))
			cmd_user(u, cmd, ac, av);
		else if (!strcasecmp(cmd, "NICK"))
			cmd_nick(u, cmd, ac, av);
/*		else if (!strcasecmp(cmd, "PONG"))
			cmd_pong(u, cmd, ac, av);
		else if (!strcasecmp(cmd, "ADMIN"))
			cmd_admin(u, cmd, ac, av);*/
		else if (!strcasecmp(cmd, "QUIT"))
		{
			cmd_quit(u, cmd, ac, av);
			rc = -1;
			goto out;
		}

		else if (!strcasecmp(cmd, "PONG"))
			cmd_noop(u, cmd, ac, av);
		else
			send_num(u, 451, cmd, ":You have not registered");	

		if ((u->flags & (UFL_HAVEUSER|UFL_HAVENICK)) == (UFL_HAVEUSER|UFL_HAVENICK))
			cmd_register(u);
	}
	else
	{
		if (!strcasecmp(cmd, "QUIT"))
		{
			cmd_quit(u, cmd, ac, av);
			rc = -1;
			goto out;
		}
		else if (!strcasecmp(cmd, "NAMES"))
			cmd_names(u, cmd, ac, av);
		else if (!strcasecmp(cmd, "JOIN"))
			cmd_join(u, cmd, ac, av);
		else if (!strcasecmp(cmd, "PRIVMSG"))
			cmd_gen_msg(u, cmd, ac, av);
		else if (!strcasecmp(cmd, "NICK"))
			cmd_nick(u, cmd, ac, av);
		else if (!strcasecmp(cmd, "PART"))
			cmd_part(u, cmd, ac, av);
		else if (!strcasecmp(cmd, "LUSERS"))
			cmd_lusers(u, cmd, ac, av);
		else if (!strcasecmp(cmd, "NOTICE"))
			cmd_gen_msg(u, cmd, ac, av);
		else if (!strcasecmp(cmd, "PING"))
			cmd_ping(u, cmd, ac, av);
		else if (!strcasecmp(cmd, "PONG"))
			cmd_noop(u, cmd, ac, av);
		else if (!strcasecmp(cmd, "MOTD"))
			send_motd(u);
/*		else if (!strcasecmp(cmd, "WHOIS"))
			cmd_whois(u, cmd, ac, av);*/
	}

	*u->inbuf = '\0';
out:
	free(source ? av - 2 : av - 1);
	free(buf);
	return(rc);
}
